package sun.reflect;

import java.security.AccessController;
import java.security.PrivilegedAction;

class MethodAccessorGenerator extends AccessorGenerator
{
    private static final short NUM_BASE_CPOOL_ENTRIES = 12;
    private static final short NUM_METHODS = 2;
    private static final short NUM_SERIALIZATION_CPOOL_ENTRIES = 2;
    private static volatile int methodSymnum;
    private static volatile int constructorSymnum;
    private static volatile int serializationConstructorSymnum;
    private Class<?> declaringClass;
    private Class<?>[] parameterTypes;
    private Class<?> returnType;
    private boolean isConstructor;
    private boolean forSerialization;
    private short targetMethodRef;
    private short invokeIdx;
    private short invokeDescriptorIdx;
    private short nonPrimitiveParametersBaseIdx;
    
    MethodAccessorGenerator() {
    }
    
    public MethodAccessor generateMethod(Class<?> clazz, String s, Class<?>[] array, Class<?> clazz2, Class<?>[] array2, int n) {
        return (MethodAccessor)this.generate(clazz, s, array, clazz2, array2, n, false, false, null);
    }
    
    public ConstructorAccessor generateConstructor(Class<?> clazz, Class<?>[] array, Class<?>[] array2, int n) {
        return (ConstructorAccessor)this.generate(clazz, "<init>", array, Void.TYPE, array2, n, true, false, null);
    }
    
    public SerializationConstructorAccessorImpl generateSerializationConstructor(Class<?> clazz, Class<?>[] array, Class<?>[] array2, int n, Class<?> clazz2) {
        return (SerializationConstructorAccessorImpl)this.generate(clazz, "<init>", array, Void.TYPE, array2, n, true, true, clazz2);
    }
    
    private MagicAccessorImpl generate(final Class<?> declaringClass, String s, Class<?>[] parameterTypes, Class<?> returnType, Class<?>[] array, int modifiers, boolean isConstructor, boolean forSerialization, Class<?> clazz) {
        ByteVector create = ByteVectorFactory.create();
        this.asm = new ClassFileAssembler(create);
        this.declaringClass = declaringClass;
        this.parameterTypes = parameterTypes;
        this.returnType = returnType;
        this.modifiers = modifiers;
        this.isConstructor = isConstructor;
        this.forSerialization = forSerialization;
        this.asm.emitMagicAndVersion();
        int n = 42;
        int usesPrimitiveTypes = this.usesPrimitiveTypes();
        if (!(usesPrimitiveTypes == 0)) {
            n = (short)(n + 72);
        }
        if (forSerialization) {
            n = (short)(n + 2);
        }
        int i = (short)(n + (short)(2 * this.numNonPrimitiveParameterTypes()));
        this.asm.emitShort(AccessorGenerator.add((short)i, (short)1));
        final String generateName = generateName(isConstructor, forSerialization);
        this.asm.emitConstantPoolUTF8(generateName);
        this.asm.emitConstantPoolClass(this.asm.cpi());
        this.thisClass = this.asm.cpi();
        if (isConstructor) {
            if (forSerialization) {
                this.asm.emitConstantPoolUTF8("sun/reflect/SerializationConstructorAccessorImpl");
            }
            else {
                this.asm.emitConstantPoolUTF8("sun/reflect/ConstructorAccessorImpl");
            }
        }
        else {
            this.asm.emitConstantPoolUTF8("sun/reflect/MethodAccessorImpl");
        }
        this.asm.emitConstantPoolClass(this.asm.cpi());
        this.superClass = this.asm.cpi();
        this.asm.emitConstantPoolUTF8(AccessorGenerator.getClassName(declaringClass, false));
        this.asm.emitConstantPoolClass(this.asm.cpi());
        this.targetClass = this.asm.cpi();
        int cpi = 0;
        if (forSerialization) {
            this.asm.emitConstantPoolUTF8(AccessorGenerator.getClassName(clazz, false));
            this.asm.emitConstantPoolClass(this.asm.cpi());
            cpi = this.asm.cpi();
        }
        this.asm.emitConstantPoolUTF8(s);
        this.asm.emitConstantPoolUTF8(this.buildInternalSignature());
        this.asm.emitConstantPoolNameAndType(AccessorGenerator.sub(this.asm.cpi(), (short)1), this.asm.cpi());
        if (this.isInterface()) {
            this.asm.emitConstantPoolInterfaceMethodref(this.targetClass, this.asm.cpi());
        }
        else if (forSerialization) {
            this.asm.emitConstantPoolMethodref((short)cpi, this.asm.cpi());
        }
        else {
            this.asm.emitConstantPoolMethodref(this.targetClass, this.asm.cpi());
        }
        this.targetMethodRef = this.asm.cpi();
        if (isConstructor) {
            this.asm.emitConstantPoolUTF8("newInstance");
        }
        else {
            this.asm.emitConstantPoolUTF8("invoke");
        }
        this.invokeIdx = this.asm.cpi();
        if (isConstructor) {
            this.asm.emitConstantPoolUTF8("([Ljava/lang/Object;)Ljava/lang/Object;");
        }
        else {
            this.asm.emitConstantPoolUTF8("(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;");
        }
        this.invokeDescriptorIdx = this.asm.cpi();
        this.nonPrimitiveParametersBaseIdx = AccessorGenerator.add(this.asm.cpi(), (short)2);
        for (int j = 0; j < parameterTypes.length; ++j) {
            Class<?> clazz2 = parameterTypes[j];
            if (!AccessorGenerator.isPrimitive(clazz2)) {
                this.asm.emitConstantPoolUTF8(AccessorGenerator.getClassName(clazz2, false));
                this.asm.emitConstantPoolClass(this.asm.cpi());
            }
        }
        this.emitCommonConstantPoolEntries();
        if (!(usesPrimitiveTypes == 0)) {
            this.emitBoxingContantPoolEntries();
        }
        if (this.asm.cpi() != i) {
            throw new InternalError("Adjust this code (cpi = " + (int)this.asm.cpi() + ", numCPEntries = " + i + ")");
        }
        this.asm.emitShort((short)1);
        this.asm.emitShort(this.thisClass);
        this.asm.emitShort(this.superClass);
        this.asm.emitShort((short)0);
        this.asm.emitShort((short)0);
        this.asm.emitShort((short)2);
        this.emitConstructor();
        this.emitInvoke();
        this.asm.emitShort((short)0);
        create.trim();
        return (MagicAccessorImpl)AccessController.doPrivileged(new PrivilegedAction<MagicAccessorImpl>() {
            
            @Override
            public MagicAccessorImpl run() {
                try {
                    return (MagicAccessorImpl)ClassDefiner.defineClass(generateName, bytes, 0, bytes.length, declaringClass.getClassLoader()).newInstance();
                }
                catch (InstantiationException | IllegalAccessException cause) {
                    throw new InternalError(cause);
                }
            }
        });
    }
    
    private void emitInvoke() {
        if (this.parameterTypes.length > 65535) {
            throw new InternalError("Can't handle more than 65535 parameters");
        }
        ClassFileAssembler classFileAssembler = new ClassFileAssembler();
        if (this.isConstructor) {
            classFileAssembler.setMaxLocals(2);
        }
        else {
            classFileAssembler.setMaxLocals(3);
        }
        int length = 0;
        if (this.isConstructor) {
            classFileAssembler.opc_new(this.targetClass);
            classFileAssembler.opc_dup();
        }
        else {
            if (AccessorGenerator.isPrimitive(this.returnType)) {
                classFileAssembler.opc_new(this.indexForPrimitiveType(this.returnType));
                classFileAssembler.opc_dup();
            }
            if (!this.isStatic()) {
                classFileAssembler.opc_aload_1();
                Label label = new Label();
                classFileAssembler.opc_ifnonnull(label);
                classFileAssembler.opc_new(this.nullPointerClass);
                classFileAssembler.opc_dup();
                classFileAssembler.opc_invokespecial(this.nullPointerCtorIdx, 0, 0);
                classFileAssembler.opc_athrow();
                label.bind();
                length = classFileAssembler.getLength();
                classFileAssembler.opc_aload_1();
                classFileAssembler.opc_checkcast(this.targetClass);
            }
        }
        Label o = new Label();
        if (this.parameterTypes.length == 0) {
            if (this.isConstructor) {
                classFileAssembler.opc_aload_1();
            }
            else {
                classFileAssembler.opc_aload_2();
            }
            classFileAssembler.opc_ifnull(o);
        }
        if (this.isConstructor) {
            classFileAssembler.opc_aload_1();
        }
        else {
            classFileAssembler.opc_aload_2();
        }
        classFileAssembler.opc_arraylength();
        classFileAssembler.opc_sipush((short)this.parameterTypes.length);
        classFileAssembler.opc_if_icmpeq(o);
        classFileAssembler.opc_new(this.illegalArgumentClass);
        classFileAssembler.opc_dup();
        classFileAssembler.opc_invokespecial(this.illegalArgumentCtorIdx, 0, 0);
        classFileAssembler.opc_athrow();
        o.bind();
        int n = this.nonPrimitiveParametersBaseIdx;
        Object o2 = null;
        int n2 = 1;
        for (int i = 0; i < this.parameterTypes.length; ++i) {
            Class<Object> clazz = this.parameterTypes[i];
            n2 = (byte)(n2 + (byte)this.typeSizeInStackSlots(clazz));
            if (o2 != null) {
                o2.bind();
                o2 = null;
            }
            if (this.isConstructor) {
                classFileAssembler.opc_aload_1();
            }
            else {
                classFileAssembler.opc_aload_2();
            }
            classFileAssembler.opc_sipush((short)i);
            classFileAssembler.opc_aaload();
            if (AccessorGenerator.isPrimitive(clazz)) {
                if (this.isConstructor) {
                    classFileAssembler.opc_astore_2();
                }
                else {
                    classFileAssembler.opc_astore_3();
                }
                Object o3 = null;
                o2 = new Label();
                for (int j = 0; j < MethodAccessorGenerator.primitiveTypes.length; ++j) {
                    Class<Object> clazz2 = MethodAccessorGenerator.primitiveTypes[j];
                    if (AccessorGenerator.canWidenTo(clazz2, clazz)) {
                        if (o3 != null) {
                            o3.bind();
                        }
                        if (this.isConstructor) {
                            classFileAssembler.opc_aload_2();
                        }
                        else {
                            classFileAssembler.opc_aload_3();
                        }
                        classFileAssembler.opc_instanceof(this.indexForPrimitiveType(clazz2));
                        o3 = new Label();
                        classFileAssembler.opc_ifeq(o3);
                        if (this.isConstructor) {
                            classFileAssembler.opc_aload_2();
                        }
                        else {
                            classFileAssembler.opc_aload_3();
                        }
                        classFileAssembler.opc_checkcast(this.indexForPrimitiveType(clazz2));
                        classFileAssembler.opc_invokevirtual(this.unboxingMethodForPrimitiveType(clazz2), 0, this.typeSizeInStackSlots(clazz2));
                        AccessorGenerator.emitWideningBytecodeForPrimitiveConversion(classFileAssembler, clazz2, clazz);
                        classFileAssembler.opc_goto(o2);
                    }
                }
                if (o3 == null) {
                    throw new InternalError("Must have found at least identity conversion");
                }
                o3.bind();
                classFileAssembler.opc_new(this.illegalArgumentClass);
                classFileAssembler.opc_dup();
                classFileAssembler.opc_invokespecial(this.illegalArgumentCtorIdx, 0, 0);
                classFileAssembler.opc_athrow();
            }
            else {
                classFileAssembler.opc_checkcast((short)n);
                n = AccessorGenerator.add((short)n, (short)2);
            }
        }
        if (o2 != null) {
            o2.bind();
        }
        int length2 = classFileAssembler.getLength();
        if (this.isConstructor) {
            classFileAssembler.opc_invokespecial(this.targetMethodRef, n2, 0);
        }
        else if (this.isStatic()) {
            classFileAssembler.opc_invokestatic(this.targetMethodRef, n2, this.typeSizeInStackSlots(this.returnType));
        }
        else if (this.isInterface()) {
            if (this.isPrivate()) {
                classFileAssembler.opc_invokespecial(this.targetMethodRef, n2, 0);
            }
            else {
                classFileAssembler.opc_invokeinterface(this.targetMethodRef, n2, (byte)n2, this.typeSizeInStackSlots(this.returnType));
            }
        }
        else {
            classFileAssembler.opc_invokevirtual(this.targetMethodRef, n2, this.typeSizeInStackSlots(this.returnType));
        }
        int length3 = classFileAssembler.getLength();
        if (!this.isConstructor) {
            if (AccessorGenerator.isPrimitive(this.returnType)) {
                classFileAssembler.opc_invokespecial(this.ctorIndexForPrimitiveType(this.returnType), this.typeSizeInStackSlots(this.returnType), 0);
            }
            else if (this.returnType == Void.TYPE) {
                classFileAssembler.opc_aconst_null();
            }
        }
        classFileAssembler.opc_areturn();
        int length4 = classFileAssembler.getLength();
        classFileAssembler.setStack(1);
        classFileAssembler.opc_invokespecial(this.toStringIdx, 0, 1);
        classFileAssembler.opc_new(this.illegalArgumentClass);
        classFileAssembler.opc_dup_x1();
        classFileAssembler.opc_swap();
        classFileAssembler.opc_invokespecial(this.illegalArgumentStringCtorIdx, 1, 0);
        classFileAssembler.opc_athrow();
        int length5 = classFileAssembler.getLength();
        classFileAssembler.setStack(1);
        classFileAssembler.opc_new(this.invocationTargetClass);
        classFileAssembler.opc_dup_x1();
        classFileAssembler.opc_swap();
        classFileAssembler.opc_invokespecial(this.invocationTargetCtorIdx, 1, 0);
        classFileAssembler.opc_athrow();
        ClassFileAssembler o4 = new ClassFileAssembler();
        o4.emitShort((short)length);
        o4.emitShort((short)length2);
        o4.emitShort((short)length4);
        o4.emitShort(this.classCastClass);
        o4.emitShort((short)length);
        o4.emitShort((short)length2);
        o4.emitShort((short)length4);
        o4.emitShort(this.nullPointerClass);
        o4.emitShort((short)length2);
        o4.emitShort((short)length3);
        o4.emitShort((short)length5);
        o4.emitShort(this.throwableClass);
        this.emitMethod(this.invokeIdx, (int)classFileAssembler.getMaxLocals(), classFileAssembler, o4, new short[] { this.invocationTargetClass });
        return;
    }
    
    private boolean usesPrimitiveTypes() {
        if (this.returnType.isPrimitive()) {
            return true;
        }
        int i = 0;
        while (i < this.parameterTypes.length) {
            if (this.parameterTypes[i].isPrimitive()) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    private int numNonPrimitiveParameterTypes() {
        int n = 0;
        for (int i = 0; i < this.parameterTypes.length; ++i) {
            if (!this.parameterTypes[i].isPrimitive()) {
                ++n;
            }
        }
        return n;
    }
    
    private boolean isInterface() {
        return this.declaringClass.isInterface();
    }
    
    private String buildInternalSignature() {
        StringBuffer sb = new StringBuffer();
        sb.append("(");
        for (int i = 0; i < this.parameterTypes.length; ++i) {
            sb.append(AccessorGenerator.getClassName(this.parameterTypes[i], true));
        }
        sb.append(")");
        sb.append(AccessorGenerator.getClassName(this.returnType, true));
        return sb.toString();
    }
    
    private static synchronized String generateName(boolean b, boolean b2) {
        if (b) {
            if (b2) {
                return "sun/reflect/GeneratedSerializationConstructorAccessor" + ++MethodAccessorGenerator.serializationConstructorSymnum;
            }
            return "sun/reflect/GeneratedConstructorAccessor" + ++MethodAccessorGenerator.constructorSymnum;
        }
        else {
            return "sun/reflect/GeneratedMethodAccessor" + ++MethodAccessorGenerator.methodSymnum;
        }
    }
    
    static {
        MethodAccessorGenerator.methodSymnum = 0;
        MethodAccessorGenerator.constructorSymnum = 0;
        MethodAccessorGenerator.serializationConstructorSymnum = 0;
    }
}
